/*
  httpservertype.c  --  ask web server for its type
  Copyright (C) 1998, 1999 Steffen Solyga <solyga@absinth.net>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#define DEBUG


#include <stdio.h>
#include <errno.h>
#include <unistd.h>		/* getopt(), read(), write(), close(), STDIN_FILENO */
#include <stdlib.h>		/* strtol(), malloc() */
#include <sys/types.h>		/* open(), socket(), connect(), ulong */
#include <sys/stat.h>		/* open() */
#include <fcntl.h>		/* open(), O_RDONLY */
#include <string.h>		/* strlen() */
#include <sys/socket.h>		/* socket(), connect() */
#include <netdb.h>		/* gethostbyname() */
#include <netinet/in.h>		/* htons(), struct_sockaddr_in, struct in_addr */
#include <arpa/inet.h>		/* inet_addr() */


#define INFO_CHANNEL	stdout		/* verbose channel */
#define	STD_CHANNEL	stdout		/* output channel */
#define HELP_CHANNEL	stdout	
#define	VERSION_CHANNEL	stdout
#define ERROR_CHANNEL	stderr
#define DEBUG_CHANNEL	stderr

#define	DEFAULT_HOST	"localhost"
#define	DEFAULT_PORT	80
#define	BUFFER_SIZE	(8*1024)
#define	SEARCH_STRING	"Server: "
#define	RETVAL_OK	0
#define	RETVAL_QUIET	1
#define	RETVAL_BUG	2
#define	RETVAL_ERROR	3

#define VERSION_NUMBER			"0.01"
#define DATE_OF_LAST_MODIFICATION	"1999/03/01"
#define	name_of_this_program		httpservertype
#define	Name_of_this_program		Httpservertype
#define	NAME_OF_THIS_PROGRAM		HTTPSERVERTYPE
#define	MANUAL_NUMBER			1
#define	MY_EMAIL_ADDRESS		"Steffen Solyga <solyga@absinth.net>"

#define	MIN( A, B )		( (A)<(B) ? (A) : (B) )

#define	N_( A )			A
/* taken from resolv/herror.c of glibc-2.0.7 */
const char *h_errlist[] = {
        N_("Resolver Error 0 (no error)"),
        N_("Unknown host"),                     /* 1 HOST_NOT_FOUND */
        N_("Host name lookup failure"),         /* 2 TRY_AGAIN */
        N_("Unknown server error"),             /* 3 NO_RECOVERY */
        N_("No address associated with name"),  /* 4 NO_ADDRESS */
};


typedef unsigned char	UCHAR;


int
dsphlp( char* pn ) {
  fprintf( HELP_CHANNEL, "%s v%s (%s)\n", pn, VERSION_NUMBER, DATE_OF_LAST_MODIFICATION );
  fprintf( HELP_CHANNEL, "Determine type of web server running on host.\n" );
  fprintf( HELP_CHANNEL, "Flowers & bug reports to %s.\n", MY_EMAIL_ADDRESS );
  fprintf( HELP_CHANNEL, "Usage: %s [options] [host]\n", pn );
  fprintf( HELP_CHANNEL, "  -h\t help, write this info to %s and exit sucessfully\n", HELP_CHANNEL==stdout?"stdout":"stderr" );
  fprintf( HELP_CHANNEL, "  -v\t raise verbosity level for %s\n", INFO_CHANNEL==stdout?"stdout":"stderr" );
  fprintf( HELP_CHANNEL, "  -V\t print version and compilation info to %s and exit sucessfully\n", VERSION_CHANNEL==stdout?"stdout":"stderr" );
  fprintf( HELP_CHANNEL, "controllers:\n");
  fprintf( HELP_CHANNEL, "  -p port\t set port (default is %d)\n", DEFAULT_PORT );
  return( 0 );
}


int
dspver( char* pn ) {
  fprintf( VERSION_CHANNEL, "%s v%s (%s)\n", pn, VERSION_NUMBER, DATE_OF_LAST_MODIFICATION );
  fprintf( VERSION_CHANNEL, "compilation settings:\n" );
  fprintf( VERSION_CHANNEL, "  DEFAULT_HOST:  %s\n", DEFAULT_HOST );
  fprintf( VERSION_CHANNEL, "  DEFAULT_PORT:  %d\n", DEFAULT_PORT );
  return( 0 );
}


int
my_recv( int s, void* buf, int len, unsigned int flags ) {
/* this is probably a subroutine without any sense */
  unsigned char* p= buf;
  int nbr;
  int tnbr= 0;
  int rem= len;
  do {
    if( (nbr=recv(s,p+tnbr,rem,flags)) == -1 ) return( -1 );
    tnbr+= nbr;
    rem-= nbr;
  } while( nbr>0 && rem>0 );
  return( tnbr );
}


ulong
get_ipaddr_n_by_name( char* hostname ) {
/* return binary address in network byte order */
  struct hostent* ent= gethostbyname( hostname );
  if( ent == NULL ) return( -1 );
  if( ent->h_addr_list[0] == NULL ) return( -1 );
  return( *( (ulong*)ent->h_addr_list[0] ) );
}


int
main( int argc, char** argv ) {
/*
SOLYGA --------------------
SOLYGA main() httpservertype

SOLYGA started      : Sat Feb 27 23:42:42 CET 1999 @beast
*/
  int retval= RETVAL_OK;
  int c;
  int verbose= 0;
  char default_hostname[128]= DEFAULT_HOST;
  char* hostname= default_hostname;
  int port= DEFAULT_PORT;
  int sockfd= 0;
  struct sockaddr_in host;
  char buffer[BUFFER_SIZE];
  char* bp= buffer;
  ulong ipaddr_n;
/***** process options *****/
  while( (c=getopt(argc,argv,"hvVp:s:")) != EOF ) {
    switch( c ) {
      case 'p': /* set port */ {
        char* p;
        port= (int)strtol( optarg, &p, 0 );
        if( p == optarg ) {
          fprintf( ERROR_CHANNEL,
                   "%s: Invalid port number `%s'.\n",
                   *argv, optarg );
          retval= RETVAL_ERROR; goto DIE_NOW;
        }
        break;
      }
      case 'h': /* display help and exit sucessfully */
        dsphlp(*argv);
        retval= RETVAL_OK; goto DIE_NOW;
      case 'v': /* raise verbosity level */
        verbose++;
        break;
      case 'V': /* display version to VERSION_CHANNEL and exit sucessfully */
        dspver( *argv );
        retval= RETVAL_OK; goto DIE_NOW;
      case '?': /* refer to -h and exit unsucessfully */
        fprintf( ERROR_CHANNEL, "Try `%s -h' for more information.\n", *argv );
        retval= RETVAL_ERROR; goto DIE_NOW;
      default : /* program error */
        fprintf( ERROR_CHANNEL, "%s: Options bug! E-mail me at %s.\n",
                 *argv, MY_EMAIL_ADDRESS );
        retval= RETVAL_BUG; goto DIE_NOW;
    }
  }

/* get hostname */
  if( optind<argc ) hostname= argv[optind];

/* open connexion */
  if( (sockfd=socket(AF_INET,SOCK_STREAM,0)) == -1 ) {
    fprintf( ERROR_CHANNEL, "%s: Couldn't create socket. %s.\n",
             *argv, sys_errlist[errno] );
    retval= RETVAL_ERROR; goto DIE_NOW;
  }
  if( verbose >= 2 ) {
    fprintf( INFO_CHANNEL, "Resolving hostname `%s'.\n", hostname );
  }
  if( (ipaddr_n=get_ipaddr_n_by_name( hostname )) == -1 ) {
    fprintf( ERROR_CHANNEL, "%s: Couldn't resolve hostname `%s'. %s.\n",
             *argv, hostname, h_errlist[h_errno] );
    retval= RETVAL_ERROR; goto DIE_NOW;
  }
  if( verbose >= 2 ) {
    char str_ipaddr[]= "000.000.000.000";
    UCHAR* p= (UCHAR*)&ipaddr_n;
    snprintf( str_ipaddr, strlen(str_ipaddr)+1, "%d.%d.%d.%d",
              p[0], p[1], p[2], p[3] );
    fprintf( INFO_CHANNEL, "IP address of `%s' is %s.\n",
             hostname, str_ipaddr );
  }

  host.sin_family= AF_INET;
  host.sin_port= htons(port);
  host.sin_addr.s_addr= ipaddr_n;
  if( verbose >= 2 ) {
    fprintf( INFO_CHANNEL,
             "Connecting to host `%s' at port %d.\n",
              hostname, port );
  }
  if( connect(sockfd,(struct sockaddr*)&host,sizeof(host)) == -1 ) {
    fprintf( ERROR_CHANNEL, "%s: Couldn't connect to host `%s' port %d. %s.\n",
             *argv, hostname, port, sys_errlist[errno] );
    retval= RETVAL_ERROR; goto DIE_NOW;
  }
/* send query */
  snprintf( buffer, BUFFER_SIZE, "%s\n%s\n%s%s\n\n",
            "HEAD / HTTP/1.0",
            "Accept: text/html",
            "User-Agent: ",
            *argv );
  if( verbose >= 3 ) {
    fprintf( INFO_CHANNEL, "Sending query:\n%s", buffer );
  }
  if( send(sockfd,buffer,BUFFER_SIZE,0) == -1 ) {
    fprintf( ERROR_CHANNEL, "%s: Couldn't send query. %s.\n",
             *argv, sys_errlist[errno] );
    retval= RETVAL_ERROR; goto DIE_NOW;
  }

/* receive answer */
  if( recv(sockfd,buffer,BUFFER_SIZE,0) == -1 ) {
    fprintf( ERROR_CHANNEL, "%s: Couldn't receive answer. %s.\n", 
             *argv, sys_errlist[errno] );
    retval= RETVAL_ERROR; goto DIE_NOW;
  }
  if( verbose >= 3 ) {
    fprintf( INFO_CHANNEL, "Received answer:\n%s", buffer );
  }

/* close connexion */
  if( close( sockfd ) == -1 ) {
    fprintf( ERROR_CHANNEL, "%s: Couldn't close socket. %s.\n",
             *argv, sys_errlist[errno] );
    retval= RETVAL_ERROR; goto DIE_NOW;
  }

/* pick out the Server type */
  if( (bp=strstr(buffer,SEARCH_STRING)) == NULL ) {
    fprintf( ERROR_CHANNEL,
             "%s: Webserver at `%s' doesn't speak about its type.\n",
             *argv, hostname );
    retval= RETVAL_QUIET; goto DIE_NOW;
  }
  bp+= strlen(SEARCH_STRING);
  { /* replace first '\n' or '\r' by '\0' */
    char* p= bp;
    while( *p != '\n' && *p != '\r' ) p++;
    *p= '\0';
  }
  if( verbose ) {
    fprintf( STD_CHANNEL, "Webserver at `%s' port %d answered `%s'.\n",
             hostname, port, bp );
  }
  else {
    fprintf( STD_CHANNEL, "%s\n", bp );
  }

DIE_NOW:
  close( sockfd );
  exit( retval );
}
